/*
 * This code is underdocumented.  Read syntaxtest.c and learn by example.
 *
 * Copyright (C) 1998,1999,2004 by Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <string.h>

#include "syntax.h"

const char *c_keywords[] = {
  "asm", "break", "case", "char", "const", "continue", "default", "do",
  "double", "else", "enum", "extern", "float", "for", "goto", "if",
  "inline", "int", "long", "register", "return", "short", "signed",
  "sizeof", "static", "struct", "switch", "typedef", "union", "unsigned",
  "void", "volatile", "while", 0
};

const char *cpp_keywords[] = {
  "asm", "break", "case", "char", "const", "continue", "default", "do",
  "double", "else", "enum", "extern", "float", "for", "goto", "if",
  "inline", "int", "long", "register", "return", "short", "signed",
  "sizeof", "static", "struct", "switch", "typedef", "union", "unsigned",
  "void", "volatile", "while",

  "auto", "catch", "class", "delete", "friend", "new", "operator",
  "private", "protected", "public", "template", "this", "throw",
  "try", "virtual", 0
};

const char *java_keywords[] = {
  "break", "case", "char", "const", "continue", "default", "do",
  "double", "else", "float", "for", "if", "int", "long", "return", "short",
  "static", "switch", "void", "volatile", "while",
 
  "catch", "class", "new", "private", "protected", "public", "this",
  "throw", "try",

  "abstract", "boolean", "byte", "extends", "final", "finally", "goto",
  "implements", "import", "instanceof", "interface", "native", "package",
  "strictfp", "super", "synchronized", "throws", "transient", "volatile",
  0

  /* one list said these were keywords, but Sun didn't:
       cast future generic inner operator outer rest var

     these ought to be treated somehow special: null, true, false

     Here are a bunch that I don't know the meanings of:
       cast future inner generic operator outer rest strictfp var volatile
  */
};

const char **keywords[] = { c_keywords, cpp_keywords, java_keywords };

#define OPERATOR_SET "(){}[]:.=!<>,;?+-*/|&^%~"
#define WHITESPACE_SET "\t "

language_t language = C;

static inline int is_op(char c) { return strchr(OPERATOR_SET, c)!=NULL; }
static inline int is_white(char c) { return strchr(WHITESPACE_SET, c)!=NULL; }
static inline int is_alpha(char c) {
  return ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '_');
}
static inline int is_num(char c, int in_hex, int dot_okay, int e_okay) {
  return (c >= '0' && c <= '9') ||
    (in_hex && ((c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f'))) ||
    (dot_okay && c == '.') ||
    (e_okay && (c == 'e' || c == 'E'));
}
static inline int is_alphanum(char c) {
  return is_alpha(c) || is_num(c, 0, 0, 0);
}
static int is_keyword(char *s, int len) {
  int i;

  for (i=0; keywords[language][i]; i++)
    if (!strncmp(s, keywords[language][i], len)
      && len == strlen(keywords[language][i]))
      return 1;

  return 0;
}

static int odd_slashes(char *p) {
  char *q = p-1;
  while (*q == '\\') q--;
  return !((p-q)%2);
}

syntax_t get_syntax(char *buf, int i, int *di, int reset) {
  static int in_comment=0;
  static int in_string=0;
  static int in_char=0;
  static int in_cpp_comment=0;
  static int in_preproc=0;
  static int in_include_string=0;
  static int no_preproc=0;

  if (reset)
    in_preproc = in_cpp_comment = no_preproc = 0;

  if (in_comment) {
    char *p = strstr(buf+i, "*/");
    if (p) {
      *di = p - &buf[i] + 2;
      in_comment = 0;
    }
    else
      *di = strlen(&buf[i]);
    return COMMENT;
  }

  if (in_string) {
    char *p = strchr(buf+i, '"');
    if (p) {
      *di = p - &buf[i] + 1;
      if (!odd_slashes(p))
        in_string = 0;
    }
    else {
      *di = strlen(&buf[i]);
      in_string = 0;      /* do strings auto-terminate on \n? */
    }
    return STRING;
  }

  if (in_char) {
    char *p = strchr(buf+i, '\'');
    if (p) {
      *di = p - &buf[i] + 1;
      if (!odd_slashes(p))
        in_char = 0;
    }
    else {
      *di = strlen(&buf[i]);
      in_char = 0;      /* do chars auto-terminate on \n? */
    }
    return CHAR;
  }

  if (in_include_string) {
    char *p = strchr(buf+i, '>');
    if (p) {
      *di = p - &buf[i] + 1;
      if (!odd_slashes(p))
        in_include_string = 0;
    }
    else {
      *di = strlen(&buf[i]);
      in_include_string = 0;      /* do #include<> strings auto-terminate
                                     on \n? */
    }
    return STRING;
  }
    
  if (!no_preproc && buf[i] == '#') {
    char *p = &buf[i + 1];

    while (is_alphanum(*p)) p++;

    *di = p - &buf[i];
    no_preproc = in_preproc = 1;
    return PREPROCESSOR;
  }
    
  if (!strncmp(&buf[i], "/*", 2)) {
    in_comment = 1;
    *di = 2;
    return COMMENT;
  }

  if ((language == CPP || language == JAVA) && !strncmp(&buf[i], "//", 2)) {
    *di = strlen(&buf[i]);
    return COMMENT;
  }

  if (buf[i] == '"') {
    no_preproc = in_string = 1;
    *di = 1;
    return STRING;
  }

  if (buf[i] == '\'') {
    no_preproc = in_char = 1;
    *di = 1;
    return CHAR;
  }

  if (in_preproc && buf[i] == '<') {
    in_include_string = 1;
    *di = 1;
    return STRING;
  }

  if (is_num(buf[i], 0, 0, 0) || (buf[i] == '.'
   && is_num(buf[i+1], 0, 0, 0))) {
    int dot_okay = buf[i]!='.', e_okay = 1;
    int in_hex = (buf[i] == '0' && buf[i+1] == 'x');
    char *p = &buf[i + (in_hex ? 2 : 1)];

    while (is_num(*p, in_hex, dot_okay, e_okay)) {
      dot_okay &= *p != '.';
      e_okay &= *p != 'e' && *p != 'E';
      p++;
    }

    *di = p - &buf[i];
    no_preproc = 1;
    return NUMBER;
  }

  if (is_alpha(buf[i])) {
    char *p = &buf[i + 1];

    while (is_alphanum(*p)) p++;

    *di = p - &buf[i];
    no_preproc = 1;
    return is_keyword(&buf[i], *di) ? KEYWORD : IDENT;
  }

  if (is_op(buf[i])) {
    *di = 1;
    no_preproc = 1;
    return OPERATOR;
  }

  if (is_white(buf[i])) {
		*di = 1;
    return WHITESPACE;
  }

  *di = 1;
  return INVALID;
}
