/* syntax - Library support for C/C++/Java syntax highlighting
 * Copyright (C) 1998,1999,2004 by Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef SYNTAX_H
#define SYNTAX_H

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
  IDENT=0, KEYWORD, COMMENT, STRING, CHAR, PREPROCESSOR, OPERATOR,
  NUMBER, INVALID, WHITESPACE, UNDEFINED, NONE
} syntax_t;

typedef enum {
  C=0, CPP, JAVA
} language_t;
extern language_t language;  /* which keyword set and comment style to use */

/* get_syntax: figure out the syntax of a string
 * parameters:
 *   char *buf: the full line in question
 *   int i:     the current index into the line
 *   int *di:   this will return how much i should be incremented
 *   int reset: this should be true each time a new line is parsed
 * return value:
 *   one of the syntax_t values defined above
 * notes:
 *   this must be called repeatedly for a line, and i must always
 *   be incremented by the amount returned in *di.  All lines are
 *   independent.  Use reset=1 when starting a new line.
 *   See chtml.c and codeviewer.c for examples.
 */
syntax_t get_syntax(char *buf, int i, int *di, int reset);

#ifdef __cplusplus
}
#endif

#endif
